"""The core of where-is. This is where the CLI frontend gets its objects from."""
import json
from typing import List, Dict, Union
from pathlib import Path
import os
from whereis import exceptions, utils
import shutil
from rich.table import Table
from rich.tabulate import tabulate_mapping


class Entry:
    def __init__(self, name: str, *locations: List[str]) -> None:
        """Initializes an Entry object.

        Args:
            name: The name of an entry.
            *locations: The locations the entry stores.
        """
        self._name = name
        self._locations = locations

    @property
    def name(self) -> str:
        """The name of the entry.

        Returns:
            The name of the entry
        """
        return self._name

    @property
    def locations(self) -> List[Path]:
        """All of the locations an entry has.

        Notes:
            The paths can be formatted.
            Here is the following format map:
                {HOME}: Your home folder.
                {WHEREIS_CONFIG}: The where-is configuration folder.
                {CONFIG_FOLDER}: The configuration folder.

        Returns:
            All of the locations an entry has.
        """
        return [
            Path(self._format_path(Path(os.path.join(os.path.sep, *location))))
            for location in self._locations
        ]

    @property
    def to_dict(self) -> Dict[str, Union[str, List[List[str]]]]:
        """Converts a entry object to a dictionary.

        Returns:
            Converted to dictionary entry object.
        """
        return {
            "name": self.name,
            "locations": list(self._locations),
        }

    @property
    def to_json(self) -> str:
        """Converts an entry object to json.

        Returns:
            Converted to json entry object.
        """
        return json.dumps(self.to_dict)

    @staticmethod
    def _format_path(path: Path) -> Path:
        """Formats a path.

        Args:
            path: The path to format.

        Returns:
            The formatted path.
        """
        format_map: Dict[str, Path] = {
            "HOME": Path().home(),
            "WHEREIS_CONFIG": utils.config_folder(),
            "CONFIG_FOLDER": utils.config_folder().parent,
        }
        try:
            path_parts: List[str] = list(path.parts)
            for index, part in enumerate(path_parts):
                for name, value in format_map.items():
                    if name in part:
                        del path_parts[index]
                        path_parts[index:index] = list(value.parts)

            return Path(*path_parts)
        except (KeyError, IndexError, ValueError):
            raise exceptions.FormatMapError(
                f"Format map not supported for path '{path}'."
            ) from None

    def locations_exists(self) -> Dict[Path, bool]:
        """Does each location exist?

        Returns:
            A dictionary of locations and whether that location exists.
        """
        return {location: location.exists() for location in self.locations}

    def __eq__(self, other) -> bool:
        try:
            return self.name == other.name and self.locations == other.locations
        except AttributeError:
            return False

    def __rich__(self) -> Table:
        """A shortcut to generate a table to generate configuration files found.

        Returns:
            A table object, usable by rich print instances.
        """
        columns: List[str] = ["Locations", "Exists", "Is File", "Is Folder"]
        table: Table = Table(title="[bold purple]Config files found")
        for column in columns:
            table.add_column(column)
        for location, exists in self.locations_exists().items():
            formatted_location: str = f"[magenta]{location}"
            formatted_exists: str = f"[red]{exists}" if not exists else f"[green4]{exists}"

            if exists and location.is_file():
                is_file: str = "[green4]True"
            elif exists and not location.is_file():
                is_file = "[red]False"
            else:
                is_file = "[red italic]Unknown"

            if exists and location.is_dir():
                is_dir: str = "[green]True"
            elif exists and not location.is_dir():
                is_dir = "[red]False"
            else:
                is_dir = "[red italic]Unknown"

            table.add_row(formatted_location, formatted_exists, is_file, is_dir)

        return table

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} object: name='{self.name}' locations={self.locations}>"


class Database:
    def __init__(self, location: Path = utils.config_folder()) -> None:
        """Initializes a Database object.

        Args:
            location: The location where the database is. Defaults to the config folder.
        """
        self._location = location

    @property
    def location(self) -> Path:
        """The location where the database is.

        Returns:
            The location where the database is.

        Raises:
            NotADirectoryError: When the database folder doesn't exist.
        """
        return self._location

    @property
    def _database(self) -> List[Dict[str, Union[str, List[List[str]]]]]:
        """All of the database entries in raw, waiting to be processed.

        Returns:
            A list of dictionaries if the file owning that dictionary's suffix is '.json'.

        Raises:
            EntryParseError: If the entry JSON can't be decoded.
        """
        ret: List[Dict[str, Union[str, List[List[str]]]]] = []
        for entry in self.location.iterdir():
            if entry.suffix == ".json":
                try:
                    ret.append(json.loads(entry.read_text()))
                except json.decoder.JSONDecodeError as error:
                    raise exceptions.EntryParseError(
                        f"Error parsing '{entry.absolute()}': {error}"
                    ) from None

        return ret

    @staticmethod
    def _entry_from_json(raw_entry: Dict[str, Union[str, List[List[str]]]]) -> Entry:
        """Converts a json dict to an Entry object.

        Args:
            raw_entry: The entry in json dict.

        Returns:
            An entry object.

        Raises:
            EntryParseError: If the raw entry doesn't follow the json schema.
        """
        try:
            return Entry(raw_entry["name"], *raw_entry["locations"])  # type: ignore
        except KeyError:
            raise exceptions.EntryParseError(
                f"JSON schema incorrect: {raw_entry}"
            ) from None

    @property
    def entries(self) -> List[Entry]:
        """A list of all entries.

        Returns:
            A list of entry objects from the database location.
        """
        return [self._entry_from_json(raw_entry) for raw_entry in self._database]

    def add(self, entry: Entry) -> None:
        """Adds an entry to the database.

        Args:
            entry: The entry object.

        Returns:
            Nothing.

        Raises:
            EntryExistsError: If the entry object exists in the database entries.
        """
        if entry in self.entries:
            raise exceptions.EntryExistsError("The database entry exists.")
        new_entry: Path = self.location / f"{entry.name}.json"
        new_entry.write_text(entry.to_json)

    def remove(self, entry: Entry) -> None:
        """Removes an entry from the database.

        Args:
            entry: The entry object.

        Returns:
            Nothing.

        Raises:
            EntryDoesNotExistError: If the entry object doesn't exist in the database entries.
        """
        if entry not in self.entries:
            raise exceptions.EntryNotFoundError("The database entry must exist.")
        entry_to_delete: Path = self.location / f"{entry.name}.json"
        entry_to_delete.unlink()

    def create(self) -> None:
        """Creates the database if it doesn't exist.

        Returns:
            Nothing.

        Raises:
            DatabaseExistsError: If the database exists.
        """
        if self.exists():
            raise exceptions.DatabaseExistsError("The database already exists!")
        sample_db: Path = Path(__file__).parent / "database"
        Path(self.location).mkdir()
        for path in sample_db.iterdir():
            shutil.copyfile(str(path), str(self.location / path.name))

    def delete(self) -> None:
        """Deletes the database if it exists.

        Notes:
            There is no way to implement a python native shortcut for this because __del__ gets called after program
            exit, which means the database will get deleted after program exit.

        Returns:
            Nothing.

        Raises:
            DatabaseNotFoundError: If the database doesn't exist.
        """
        if not self.location.exists():
            raise exceptions.DatabaseNotFoundError("The database doesn't exist!")
        return shutil.rmtree(self.location)

    def exists(self) -> bool:
        """Check if the database exists.

        Returns:
            True if the folder contacting the database exists, else False
        """
        return self.location.exists()

    def __rich__(self) -> Table:
        """A shortcut to generate database info.

        Returns:
            A table, usable by rich print instances.
        """
        map_: Dict[str, Union[Path, List[Entry], bool]] = {
            "Location": self.location,
            "Entries": self.entries,
            "Exists": self.exists(),
        }
        table: Table = tabulate_mapping(map_, title="Database Info")

        return table

    def __add__(self, other: Entry):
        """A python native shortcut of database.add().

        Args:
            other: The other class (should be an Entry object)

        Returns:
            Database: The modified database.
        """
        self.add(other)
        return self

    def __sub__(self, other: Entry):
        """A python native shortcut of database.remove().

        Args:
            other: The other class (should be an Entry object)

        Returns:
            Database: The modified database.
        """
        self.remove(other)
        return self

    def __enter__(self):
        """Create a temporary database. Useful for testing.

        Returns:
            Database: A database object.
        """
        self.create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Deletes a database object after context manager exit. Useful for testing.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The traceback.

        Returns:
            Nothing.
        """
        self.delete()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} object: location={self.location}>"



"""The main entry point for pip and `python -m`."""
from typing import List
import sys


def process_imports(package_name: str) -> bool:
    """Processes imports to determine if they're installed or not.

    Args:
        package_name: The package name.

    Returns:
        True if the package is installed, else False.
    """
    try:
        __import__(package_name)
        return True
    except ImportError:
        print(
            f" [✗] The package {package_name} isn't installed on your system.",
            file=sys.stderr,
        )
        return False


def main() -> None:
    """Main entry point.

    Returns:
        Nothing.
    """
    to_import: List[str] = ["rich", "typer"]
    if False in (process_imports(package_name) for package_name in to_import):
        sys.exit(2)

    from whereis.cli import main as cli_main

    return cli_main()


if __name__ == "__main__":
    main()



"""Exceptions used by where-is."""


class WhereIsException(Exception):
    """Base class for all where-is related exceptions."""

    @property
    def message(self) -> str:
        """Gets error message.

        Returns:
            The error message.
        """
        return self.args[0]

    @property
    def name(self) -> str:
        """Gets the error name.

        Returns:
            The error name.
        """
        return self.__class__.__name__


class EntryExistsError(WhereIsException):
    """Raised when a database entry exists."""


class EntryNotFoundError(WhereIsException):
    """Raised when a database entry doesn't exist."""


class FormatMapError(WhereIsException):
    """Raised when a format map is invalid."""


class EntryParseError(WhereIsException):
    """Raised when an error is encountered during parsing of a json entry."""


class DatabaseExistsError(WhereIsException):
    """Raised when a database exists."""


class DatabaseNotFoundError(WhereIsException):
    """Raised when a database isn't found."""



"""Custom messages."""
from typing import Literal, Dict
from rich.console import Console
import sys

_console: Console = Console(file=sys.stderr)


def _levels(
    message: str,
    level: Literal["info", "success", "warn", "error"],
    no_icon: bool = False,
) -> None:
    """An easier way to get levels.

    Args:
        message: The message.
        level: The level.
        no_icon: Does the message have an icon or not?

    Returns:
        Nothing.
    """
    to_message: Dict[str, str] = {
        "info": " [bold][dark_blue]{icon} [blue]{line}",
        "success": " [bold][dark_green]{icon} [green4]{line}",
        "warn": " [bold][yellow]{icon} [yellow3]{line}",
        "error": " [bold][dark_red]{icon} [red]{line}",
    }
    icon: Dict[str, str] = {"info": "🛈", "success": "✓", "warn": "⚠", "error": "✗"}

    for line in message.splitlines():
        _console.print(
            to_message[level].format(
                icon="" if no_icon else f"[[{icon[level]}]]", line=line
            )
        )


def info(message: str, no_icon: bool = False) -> None:
    """Prints out an info message.

    Args:
        message: The message.
        no_icon: Does the message have an icon or not?

    Returns:
        Nothing.
    """
    return _levels(message, "info", no_icon)


def success(message: str, no_icon: bool = False) -> None:
    """Prints out a success message.

    Args:
        message: The message.
        no_icon: Does the message have an icon or not?

    Returns:
        Nothing.
    """
    return _levels(message, "success", no_icon)


def warn(message: str, no_icon: bool = False) -> None:
    """Prints out a warning message.

    Args:
        message: The message.
        no_icon: Does the message have an icon or not?

    Returns:
        Nothing.
    """
    return _levels(message, "warn", no_icon)


def error(message: str, no_icon: bool = False) -> None:
    """Prints out an error message.

    Args:
        message: The message.
        no_icon: Does the message have an icon or not?

    Returns:
        Nothing.
    """
    return _levels(message, "error", no_icon)


def debug(message: str) -> None:
    """Prints out a debug message.

    Args:
        message: The message.

    Returns:
        Nothing.
    """
    for line in message.splitlines():
        _console.log(f"[cyan]🔍 {line}")


def _test_levels() -> None:
    """Tests levels.

    Returns:
        Nothing.
    """
    levels = [info, success, warn, error, debug]
    for level in levels:
        level("Hello, World!")  # type: ignore
        try:
            level("No icon!", no_icon=True)  # type: ignore
        except TypeError:
            pass


if __name__ == "__main__":
    _test_levels()



# where-is: Finds config files.
# Copyright (C) 2020 ALinuxPerson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The where-is package itself."""
from whereis.core import *
from whereis import levels
from rich.traceback import Traceback, TracebackType
from rich.console import Console
from typing import Type, Text
import sys
from whereis.__version__ import __version__ as version


def excepthook(
    type_: Type[BaseException], value: BaseException, traceback: TracebackType,
) -> None:
    """The except hook used globally by where-is.

    Args:
        type_: The exception type.
        value: The exception value.
        traceback: The traceback object.

    Returns:
        Nothing.
    """
    levels.error(f"[bold]Exception[/] occurred, please wait for it to be processed...")
    traceback_console: Console = Console(file=sys.stderr)

    traceback_console.print(
        Traceback.from_exception(
            type_,
            value,
            traceback,
            width=100,
            extra_lines=3,
            theme=None,
            word_wrap=False,
        )
    )
    levels.error(
        f"Please report this exception here: [blue underline]https://github.com/what-to-code-complete/where-is/issues"
    )
    levels.info(
        "[bold]Post-mortem information[/]\n"
        "[underline]How to report this exception:[/]\n"
        "[italic]1. Capture stdout and stderr to file:\n"
        "\t$ <YOUR_ERRORED_CMD> >> error.log 2>&1\n"
        "[italic]2. Upload 'error.log' as a gist [bold](note: only gists are allowed.)[/]\n"
        "[italic]3. Attach it to your issue [bold](give your issue a descriptive title and description!)[/]"
    )


sys.excepthook = excepthook


def input(*prompt: Union[str, Text], markup: bool = True, emoji: bool = True) -> str:
    """An input function that supports rich syntax.

    Args:
        *prompt: The text you want to print out to the user.
        markup: Enable markup?
        emoji: Enable emojis?

    Returns:
        The text given out by the user.
    """
    console: Console = Console()
    return console.input(*prompt, markup=markup, emoji=emoji)



"""Some useful utilities to be used by where-is."""
from pathlib import Path
import platform
from typing import Dict
import os


def config_folder(system: str = platform.system()) -> Path:
    """Gets the config folder of each operating system.

    Args:
        system: The operating system to retrieve a config folder from.

    Returns:
        A path object that points to where a config folder is
        (if system is not in Linux, Mac, Windows it will default to Linux)
    """
    switch_case: Dict[str, Path] = {
        "Linux": Path().home() / ".config" / "where-is",
        "Mac": Path().home() / "Library" / "Preferences" / "where-is",
        "Windows": Path(
            str(os.getenv("APPDATA"))  # in case the os is other than windows
        )
        / "where-is",
    }

    return switch_case.get(system, switch_case["Linux"])



"""The cli frontend for where-is."""
import typer
from pathlib import Path
from whereis import utils, levels, Database, Entry, input, version, exceptions
from typing import Optional, List
from rich import print
from rich.console import Console

app: typer.Typer = typer.Typer(
    help="An elegant way to find configuration files (and folders)."
)
is_verbose: bool = False
database_location: Path = utils.config_folder()
VERSION_STRING: str = f"""[bold dark_blue]  ---       [/][italic]where-is[/] {version} Copyright (C) 2020
[bold dark_blue] /          [/]Made by [italic bold]ALinuxPerson[/]. This project uses the [italic bold]GNU GPLv3[/] license.
[bold dark_blue]<  ?
 \\          [/][italic]This program comes with [bold]ABSOLUTELY NO WARRANTY[/]; This is free software,
[bold dark_blue]  ---       [/]and you are welcome to redistribute it under certain conditions."""


def _log(message: str) -> None:
    return levels.debug(message) if is_verbose else None


def _get_entry(
    entry_name: str, database: Database, no_err: bool = False
) -> Optional[Entry]:
    """Gets an entry safely.

    This function prevents exceptions from being printed out from core and instead replaces them with not so verbose,
    user friendly messages.

    Args:
        entry_name: The entry name.
        database: The database.
        no_err: Should no errors be displayed?

    Returns:
        An entry if no error was encountered, else nothing.
    """
    for entry_ in database.entries:
        _log(f"Checking if [bold]'{entry_.name}'[/] == [bold]'{entry_name}'[/]...")
        try:
            if entry_.name == entry_name:
                _log(f"Got entry, {entry_}")
                return entry_
        except exceptions.FormatMapError as error:
            levels.error(f"Entry formatting error: [italic]{error.message}")
            return None
    else:
        if not no_err:
            levels.error(f"Couldn't find entry '{entry_name}' in the database.")
        else:
            _log(
                f"Couldn't find any entry, but the no_err argument is True, so not printing any errors."
            )
        return None


def _eval_db_opts(info: bool, add: bool, remove: bool, delete: bool) -> bool:
    """Evaluates the options given by the user in the `where-is database [OPTIONS]` argument.

    The `where-is database [OPTIONS]` arguments are all mutually exclusive, which means this needs to be implemented to
    enforce that mutually-exclusiveness.

    Args:
        info: The info option.
        add: The add option.
        remove: The remove option.
        delete: The delete option.

    Returns:
        True if the options are mutually exclusive, else False.
    """
    _log(
        f"Got options:\n"
        f"info: {info}, add: {add}, remove: {remove}, delete: {delete}"
    )
    opts: List[bool] = [info, add, remove, delete]
    if opts.count(True) > 1:
        levels.error(
            f"The info, add, remove and delete options are mutually exclusive with each other."
        )
        return False
    return True


def _get_database(location: Path) -> Optional[Database]:
    """Gets a database object safely.

    This function prevents exceptions from being printed out from core and instead replaces them with not so verbose,
    user friendly messages. If the database doesn't exist, this function creates it automatically.

    Args:
        location: The location where the database is.

    Returns:
        A database object if no error was encountered, else nothing.
    """
    database: Database = Database(location)
    if not database.exists():
        try:
            levels.info("Database doesn't exist, creating...")
            database.create()
        except PermissionError as error:
            levels.error(
                f"Can't create database at location '{location}': [italic]{error}"
            )
            return None
    try:
        _: List[Entry] = database.entries
    except exceptions.EntryParseError as error:
        levels.error(f"Database error: [italic]{error.message}")
        return None
    _log(f"Got database, {database}")

    return database


def _add_entry(database: Database) -> bool:
    """Adds an entry safely.

    This function helps the user to add an entry to a database.

    Args:
        database: The database object.

    Returns:
        True if no error was encountered, else False.
    """
    levels.info("Enter the name of the entry.")
    entry_name: str = input("[blue]Entry name: ")
    if entry_name in [entry.name for entry in database.entries]:
        levels.error("That entry already exists.")
        return False
    entry_locations: List[str] = []
    levels.info("Enter locations for the entry. Press ctrl-C to finish.")
    while True:
        try:
            entry_location: str = input("[blue]Entry locations: ")
            entry_locations.append(entry_location)
        except KeyboardInterrupt:
            break
    entry: Entry = Entry(entry_name, *[Path(entry_).parts for entry_ in entry_locations])  # type: ignore
    database += entry  # type: ignore
    levels.success("Added entry to database.")
    return True


def _rm_entry(database: Database) -> None:
    """Removes an entry safely.

    This function prevents exceptions from being printed out from core and instead replaces them with not so verbose,
    user friendly messages.

    Args:
        database: The database object.

    Returns:
        Nothing.
    """
    levels.info("Enter the name of the entry: ")
    entry: Optional[Entry] = _get_entry(input("[blue]Entry name: "), database)
    if not entry:
        return
    database -= entry  # type: ignore
    levels.success("Removed entry from the database.")


def _del_db(database: Database) -> None:
    """Deletes a database safely.

    This function prevents exceptions from being printed out from core and instead replaces them with not so verbose,
    user friendly messages.

    Args:
        database: The database object.

    Returns:
        Nothing.
    """
    try:
        database.delete()
        levels.success("Successfully deleted database.")
    except exceptions.DatabaseNotFoundError as error:
        levels.error(f"Error: [italic]{error}")


def _show_version(value: bool) -> None:
    """Shows the formatted version.

    Args:
        value: Required otherwise _show_version always gets executed.

    Returns:
        Nothing.
    """
    if value:
        console: Console = Console()
        console.print(VERSION_STRING, style="blue")
        raise typer.Exit()


# noinspection PyUnusedLocal
@app.callback()
def root(
    verbose: bool = typer.Option(False, help="Enable verbose output."),
    version_: bool = typer.Option(
        None,
        "--version",
        help="Show this program's version number and credits",
        callback=_show_version,
    ),
    database_location_: Path = typer.Option(
        None, "--database-location", help="Specify the database location."
    ),
) -> None:
    """The root arguments.

    Args:
        verbose: Enable verbose output.
        version_: Show version.
        database_location_: Specify the database location.

    Returns:
        Nothing.
    """
    global is_verbose
    global database_location

    if verbose:
        is_verbose = True

    if database_location_:
        database_location = database_location_


@app.command()
def find(name: str = typer.Argument(..., help="The name of the entry.")) -> None:
    """Find an entry with the name NAME"""
    database: Optional[Database] = _get_database(database_location)
    if not database:
        return
    entry_: Optional[Entry] = _get_entry(name, database)
    if not entry_:
        return
    print(entry_)


@app.command("database")
def cli_database(
    info: bool = typer.Option(False, "--info", help="Show information about an entry."),
    add: bool = typer.Option(False, "--add", help="Add an entry to a database."),
    remove: bool = typer.Option(
        False, "--remove", help="Remove an entry from a database."
    ),
    delete: bool = typer.Option(False, "--delete", help="Deletes the database."),
) -> None:
    """Query, add and remove entries from the database and perform operations on the database itself."""
    database: Optional[Database] = _get_database(
        database_location
    ) if not delete else Database(database_location)
    if not _eval_db_opts(info, add, remove, delete) or not database:
        return
    if info:
        try:
            return print(database)
        except exceptions.FormatMapError as error:
            levels.error(
                f"Unable to create table:\n" f"Entry formatting error: [italic]{error}"
            )
    elif add:
        _add_entry(database)
    elif remove:
        _rm_entry(database)
    elif delete:
        _del_db(database)
    else:
        levels.info(
            "What do you want to do? pass the [bold]'--help'[/] argument to get help."
        )


def main() -> None:
    """The main entry point.

    Returns:
        Nothing.
    """
    return app()



from typing import List, Dict
import setuptools  # type: ignore
import pathlib

REQUIRED: List[str] = ["rich==3.1.0", "typer==0.3.0"]
EXTRAS: Dict[str, List[str]] = {"dev": ["mypy==0.782", "pytest==5.4.3"]}

setuptools.setup(
    name="where-is",
    version=pathlib.Path("VERSION").read_text().strip(),
    description="Finds config files.",
    long_description_content_type="text/markdown",
    long_description=pathlib.Path("README.md").read_text().strip(),
    author="ALinuxPerson",
    author_email="micheal02052007@gmail.com",
    python_requires=">=3.6.0",
    packages=setuptools.find_packages(
        exclude=["tests", "*.tests", "*.tests.*", "tests.*"]
    ),
    entry_points={"console_scripts": ["where-is=whereis.__main__:main"]},
    include_package_data=True,
    license="GNU GPLv3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)
